#
# Copyright (c) 2024 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#
import dataclasses as dc
import enum
from typing import List, Optional, Tuple, Union

from devicetree import edtlib

from nrfregtool.common import AddressOffset, NrfPsel, Product, log_dbg


@dc.dataclass(frozen=True)
class GpiosProp:
    """CTRLSEL lookup table entry for *-gpios properties"""

    name: str
    port: int
    pin: int


@enum.unique
class Ctrlsel(enum.IntEnum):
    """
    Enumeration of GPIO.PIN_CNF[n].CTRLSEL values.
    The list here may not be exhaustive.
    """

    GPIO = 0
    VPR_GRC = 1
    CAN_PWM_I3C = 2
    SERIAL0 = 3
    EXMIF_RADIO_SERIAL1 = 4
    CAN_TDM_SERIAL2 = 5
    CAN = 6
    TND = 7


# Default CTRLSEL value indicating that CTRLSEL should not be used
CTRLSEL_DEFAULT = Ctrlsel.GPIO

# Pin functions used with pinctrl, see include/zephyr/dt-bindings/pinctrl/nrf-pinctrl.h
# Only the functions relevant for CTRLSEL deduction have been included.
NRF_FUN_UART_TX = 0
NRF_FUN_UART_RX = 1
NRF_FUN_UART_RTS = 2
NRF_FUN_UART_CTS = 3
NRF_FUN_SPIM_SCK = 4
NRF_FUN_SPIM_MOSI = 5
NRF_FUN_SPIM_MISO = 6
NRF_FUN_SPIS_SCK = 7
NRF_FUN_SPIS_MOSI = 8
NRF_FUN_SPIS_MISO = 9
NRF_FUN_SPIS_CSN = 10
NRF_FUN_TWIM_SCL = 11
NRF_FUN_TWIM_SDA = 12
NRF_FUN_PWM_OUT0 = 22
NRF_FUN_PWM_OUT1 = 23
NRF_FUN_PWM_OUT2 = 24
NRF_FUN_PWM_OUT3 = 25
NRF_FUN_EXMIF_CK = 35
NRF_FUN_EXMIF_DQ0 = 36
NRF_FUN_EXMIF_DQ1 = 37
NRF_FUN_EXMIF_DQ2 = 38
NRF_FUN_EXMIF_DQ3 = 39
NRF_FUN_EXMIF_DQ4 = 40
NRF_FUN_EXMIF_DQ5 = 41
NRF_FUN_EXMIF_DQ6 = 42
NRF_FUN_EXMIF_DQ7 = 43
NRF_FUN_EXMIF_CS0 = 44
NRF_FUN_EXMIF_CS1 = 45
NRF_FUN_CAN_TX = 46
NRF_FUN_CAN_RX = 47

NRF_FUN_TDM_SCK_M = 48
NRF_FUN_TDM_SCK_S = 49
NRF_FUN_TDM_FSYNC_M = 50
NRF_FUN_TDM_FSYNC_S = 51
NRF_FUN_TDM_SDIN = 52
NRF_FUN_TDM_SDOUT = 53
NRF_FUN_TDM_MCK = 54

# Deliberately defined as placeholders
NRF_FUN_EXMIF_RWDS = 0xFF - 2
NRF_FUN_I3C_SDA = 0xFF - 1
NRF_FUN_I3C_SCL = 0xFF

_PINCTRL_CTRLSEL_LOOKUP_NRF54H20 = {
    # I3C120
    0x5F8D_3000: {
        # P2
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=0): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=6, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=6, pin=1): Ctrlsel.CAN_PWM_I3C,
    },
    # CAN120
    0x5F8D_8000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=9): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=8): Ctrlsel.CAN_PWM_I3C,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=5): Ctrlsel.CAN,
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=4): Ctrlsel.CAN,
    },
    # I3C121
    0x5F8D_E000: {
        # P2
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=2): Ctrlsel.CAN_PWM_I3C,
        # P7
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=7, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=7, pin=2): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM120
    0x5F8E_4000: {
        # P2
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=2, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=2, pin=5): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=2, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=2, pin=7): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=7): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=9): Ctrlsel.CAN_PWM_I3C,
        # P7
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=7, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=7, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=7, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=7, pin=7): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM130
    0x5F9A_4000: {
        # P9
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=9, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=9, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=9, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=9, pin=5): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIM130/SPIS130/TWIM130/TWIS130/UARTE130
    0x5F9A_5000: {
        # SPIM mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=4): Ctrlsel.SERIAL0,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=4): Ctrlsel.SERIAL0,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=4): Ctrlsel.SERIAL0,
        # UARTÈ mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=3): Ctrlsel.SERIAL0,
    },
    # SPIM131/SPIS131/TWIM131/TWIS131/UARTE131
    0x5F9A_6000: {
        # SPIM mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        GpiosProp(name="cs-gpios", port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        # UARTÈ mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # SPIS120
    0x5F8E_5000: {
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=6, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=6, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=6, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=6, pin=0): Ctrlsel.SERIAL0,
    },
    # SPIM120/UARTE120
    0x5F8E_6000: {
        # SPIM P6 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=7): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=1): Ctrlsel.SERIAL0,
        # SPIM P7 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=7, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=7, pin=6): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=7, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=7, pin=3): Ctrlsel.SERIAL0,
        # SPIM P2 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=5): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=2, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=3): Ctrlsel.SERIAL0,
        # UARTÈ P6 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=6, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=6, pin=5): Ctrlsel.SERIAL0,
        # UARTÈ P7 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=7, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=7, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=7, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=7, pin=5): Ctrlsel.SERIAL0,
        # UARTÈ P2 mappings
        NrfPsel(fun=NRF_FUN_UART_TX, port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_CTS, port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=2, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=2, pin=7): Ctrlsel.SERIAL0,
    },
    # SPIM121
    0x5F8E_7000: {
        # SPIM P6 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=13): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=12): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=6, pin=10): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=2): Ctrlsel.SERIAL0,
        # SPIM P7 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=7, pin=1): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=7, pin=0): Ctrlsel.EXMIF_RADIO_SERIAL1,
        GpiosProp(name="cs-gpios", port=7, pin=4): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=7, pin=2): Ctrlsel.EXMIF_RADIO_SERIAL1,
        # SPIM P2 mappings
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=11): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=10): Ctrlsel.SERIAL0,
        GpiosProp(name="cs-gpios", port=2, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=2): Ctrlsel.SERIAL0,
    },
    # EXMIF
    0x5F09_5000: {
        NrfPsel(fun=NRF_FUN_EXMIF_CK, port=6, pin=0): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_RWDS, port=6, pin=2): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_CS0, port=6, pin=3): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ7, port=6, pin=4): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ1, port=6, pin=5): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ6, port=6, pin=6): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ0, port=6, pin=7): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ5, port=6, pin=8): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ3, port=6, pin=9): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ2, port=6, pin=10): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_DQ4, port=6, pin=11): Ctrlsel.EXMIF_RADIO_SERIAL1,
        NrfPsel(fun=NRF_FUN_EXMIF_CS1, port=6, pin=13): Ctrlsel.EXMIF_RADIO_SERIAL1,
    },
    # TDM130
    0x5F99_2000: {
        # TDM P1 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=1, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=1, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=1, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=1, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=1, pin=5): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=1, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=1, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        # TDM P2 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=2, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=2, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=2, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=2, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=2, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # TDM131
    0x5F99_7000: {
        # TDM P1 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=1, pin=0): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=1, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=1, pin=1): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=1, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=1, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=1, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=1, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        # TDM P2 mappings
        NrfPsel(fun=NRF_FUN_TDM_MCK, port=2, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_M, port=2, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SCK_S, port=2, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDOUT, port=2, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_SDIN, port=2, pin=6): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_M, port=2, pin=7): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_TDM_FSYNC_S, port=2, pin=7): Ctrlsel.CAN_TDM_SERIAL2,
    },
}

_PINCTRL_CTRLSEL_LOOKUP_NRF9280 = {
    # I3C120
    0x5F8D_3000: {
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=9): Ctrlsel.CAN_PWM_I3C,
    },
    # CAN120
    0x5F8D_8000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=10): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=11): Ctrlsel.CAN_TDM_SERIAL2,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=4): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=5): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # CAN121
    0x5F8D_B000: {
        # P2
        NrfPsel(fun=NRF_FUN_CAN_RX, port=2, pin=8): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=2, pin=9): Ctrlsel.CAN_TDM_SERIAL2,
        # P9
        NrfPsel(fun=NRF_FUN_CAN_RX, port=9, pin=2): Ctrlsel.CAN_TDM_SERIAL2,
        NrfPsel(fun=NRF_FUN_CAN_TX, port=9, pin=3): Ctrlsel.CAN_TDM_SERIAL2,
    },
    # I3C121
    0x5F8D_E000: {
        NrfPsel(fun=NRF_FUN_I3C_SCL, port=2, pin=10): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_I3C_SDA, port=2, pin=11): Ctrlsel.CAN_PWM_I3C,
    },
    # PWM120
    0x5F8E_4000: {
        # P2
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=2, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=2, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=2, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=2, pin=3): Ctrlsel.CAN_PWM_I3C,
        # P6
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=0): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=6, pin=6): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=1): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=6, pin=7): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=6, pin=8): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=6, pin=9): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIS120
    0x5F8E_5000: {
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=6, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=6, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=6, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=6, pin=0): Ctrlsel.SERIAL0,
    },
    # SPIM120/UARTE120
    0x5F8E_6000: {
        # SPIM P2 mappings
        GpiosProp(name="cs-gpios", port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=0): Ctrlsel.SERIAL0,
        # SPIM P6 mappings
        GpiosProp(name="cs-gpios", port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=1): Ctrlsel.SERIAL0,
        # UARTE P2 mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=2, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=2, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=2, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=2, pin=4): Ctrlsel.SERIAL0,
        # UARTE P6 mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=6, pin=7): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=6, pin=5): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=6, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=6, pin=8): Ctrlsel.SERIAL0,
    },
    # SPIM121
    0x5F8E_7000: {
        # P2
        GpiosProp(name="cs-gpios", port=2, pin=6): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=2, pin=8): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=2, pin=9): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=2, pin=1): Ctrlsel.SERIAL0,
        # P6
        GpiosProp(name="cs-gpios", port=6, pin=10): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=6, pin=12): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=6, pin=13): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=6, pin=2): Ctrlsel.SERIAL0,
    },
    # PWM130
    0x5F9A_4000: {
        NrfPsel(fun=NRF_FUN_PWM_OUT0, port=9, pin=2): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT1, port=9, pin=3): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT2, port=9, pin=4): Ctrlsel.CAN_PWM_I3C,
        NrfPsel(fun=NRF_FUN_PWM_OUT3, port=9, pin=5): Ctrlsel.CAN_PWM_I3C,
    },
    # SPIM130/SPIS130/TWIM130/TWIS130/UARTE130
    0x5F9A_5000: {
        # SPIM mappings
        GpiosProp(name="cs-gpios", port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MISO, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_MOSI, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIM_SCK, port=9, pin=0): Ctrlsel.SERIAL0,
        # SPIS mappings
        NrfPsel(fun=NRF_FUN_SPIS_CSN, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MISO, port=9, pin=3): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_MOSI, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_SPIS_SCK, port=9, pin=0): Ctrlsel.SERIAL0,
        # TWIM mappings
        NrfPsel(fun=NRF_FUN_TWIM_SCL, port=9, pin=0): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_TWIM_SDA, port=9, pin=3): Ctrlsel.SERIAL0,
        # UARTE mappings
        NrfPsel(fun=NRF_FUN_UART_CTS, port=9, pin=2): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RTS, port=9, pin=1): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_RX, port=9, pin=4): Ctrlsel.SERIAL0,
        NrfPsel(fun=NRF_FUN_UART_TX, port=9, pin=3): Ctrlsel.SERIAL0,
    },
}

_PINCTRL_CTRLSEL_LOOKUP = {
    Product.NRF54H20: _PINCTRL_CTRLSEL_LOOKUP_NRF54H20,
    Product.NRF9280: _PINCTRL_CTRLSEL_LOOKUP_NRF9280,
}


def dt_lookup_ctrlsel(
    product: Product,
    src: Union[edtlib.PinCtrl, edtlib.Property],
    psel: Union[NrfPsel, Tuple[int, int]],
) -> Optional[int]:
    """Get the CTRLSEL value to use for a given pin selection."""

    lut = _PINCTRL_CTRLSEL_LOOKUP[product]

    if isinstance(src, edtlib.PinCtrl):
        identifier = src.node.regs[0].addr
        sub_entry = psel
    elif isinstance(src, edtlib.Property):
        try:
            identifier = src.node.regs[0].addr
        except IndexError:
            identifier = src.node.label
        sub_entry = GpiosProp(name=src.name, port=psel[0], pin=psel[1])
    else:
        raise ValueError(f"Unsupported GPIO pin source: {src}")

    if identifier in lut:
        ctrlsel = lut[identifier].get(sub_entry, CTRLSEL_DEFAULT)
    else:
        ctrlsel = None

    log_dbg(
        f"identifier={hex(identifier) if isinstance(identifier, int) else identifier}, "
        f"{sub_entry=} -> {ctrlsel=}"
    )

    return ctrlsel


NRF_SAADC_AIN0 = 1
NRF_SAADC_AIN1 = 2
NRF_SAADC_AIN2 = 3
NRF_SAADC_AIN3 = 4
NRF_SAADC_AIN4 = 5
NRF_SAADC_AIN5 = 6
NRF_SAADC_AIN6 = 7
NRF_SAADC_AIN7 = 8

_CHANNEL_LOOKUP = {
    # SAADC
    0x5F98_2000: {
        NRF_SAADC_AIN0: (1, 0),
        NRF_SAADC_AIN1: (1, 1),
        NRF_SAADC_AIN2: (1, 2),
        NRF_SAADC_AIN3: (1, 3),
        NRF_SAADC_AIN4: (1, 4),
        NRF_SAADC_AIN5: (1, 5),
        NRF_SAADC_AIN6: (1, 6),
        NRF_SAADC_AIN7: (1, 7),
    }
}


def dt_lookup_adc_channel_pins(channel: edtlib.Node) -> List[Tuple[int, int]]:
    lut = _CHANNEL_LOOKUP

    address = channel.parent.regs[0].addr | (1 << AddressOffset.SECURITY)
    if address not in lut:
        return []

    pins = []

    if "zephyr,input-positive" in channel.props:
        positive_pin = lut[address].get(channel.props["zephyr,input-positive"].val)
        if positive_pin is not None:
            log_dbg(
                f"{address=:08x}, {channel.props['zephyr,input-positive']=} -> {positive_pin=}"
            )

            pins.append(positive_pin)

    if "zephyr,input-negative" in channel.props:
        negative_pin = lut[address].get(channel.props["zephyr,input-negative"].val)
        if negative_pin is not None:
            log_dbg(
                f"{address=:08x}, {channel.props['zephyr,input-negative']=} -> {negative_pin=}"
            )

            pins.append(negative_pin)

    return pins


NRF_COMP_LPCOMP_AIN0 = "AIN0"
NRF_COMP_LPCOMP_AIN1 = "AIN1"
NRF_COMP_LPCOMP_AIN2 = "AIN2"
NRF_COMP_LPCOMP_AIN3 = "AIN3"
NRF_COMP_LPCOMP_AIN4 = "AIN4"
NRF_COMP_LPCOMP_AIN5 = "AIN5"
NRF_COMP_LPCOMP_AIN6 = "AIN6"
NRF_COMP_LPCOMP_AIN7 = "AIN7"

_COMP_LPCOMP_LOOKUP = {
    # COMP/LPCOMP
    0x5F98_3000: {
        NRF_COMP_LPCOMP_AIN0: (1, 0),
        NRF_COMP_LPCOMP_AIN1: (1, 1),
        NRF_COMP_LPCOMP_AIN2: (1, 2),
        NRF_COMP_LPCOMP_AIN3: (1, 3),
        NRF_COMP_LPCOMP_AIN4: (1, 4),
        NRF_COMP_LPCOMP_AIN5: (1, 5),
        NRF_COMP_LPCOMP_AIN6: (1, 6),
        NRF_COMP_LPCOMP_AIN7: (1, 7),
    }
}


def dt_lookup_comp_lpcomp_pins(comp_lpcomp: edtlib.Node) -> List[Tuple[int, int]]:
    lut = _COMP_LPCOMP_LOOKUP

    address = comp_lpcomp.regs[0].addr | (1 << AddressOffset.SECURITY)
    if address not in lut:
        return []

    pins = []

    if "psel" in comp_lpcomp.props:
        psel = lut[address].get(comp_lpcomp.props["psel"].val)
        if psel:
            log_dbg(f"{address=:08x}, psel -> {psel=}")
            pins.append(psel)

    if "extrefsel" in comp_lpcomp.props:
        extrefsel = lut[address].get(comp_lpcomp.props["extrefsel"].val)
        if extrefsel:
            log_dbg(f"{address=:08x}, extrefsel -> {extrefsel=}")
            pins.append(extrefsel)

    return pins


@dc.dataclass
class HardwareId:
    part_code: int
    revision: int


_HARDWARE_ID_LOOKUP = {
    Product.NRF54H20: HardwareId(part_code=0x16, revision=0x2),
    Product.NRF9280: HardwareId(part_code=0x12, revision=0x1),
}


def lookup_hardware_id(product: Product) -> HardwareId:
    """
    Look up hardware ID information based on a product.

    :param product: Product.

    :return: Hardware ID information.
    """
    return _HARDWARE_ID_LOOKUP[product]



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

import dataclasses as dc
from collections import defaultdict
from functools import cached_property
from pprint import pformat
from typing import List, NamedTuple, Optional, Union

from devicetree import edtlib

from .common import Record, log_dbg, log_dev, log_vrb
from .parsed_dt import (
    ExtractedConfig,
    InterpretedPeripheral,
    ParsedDTNode,
    PeripheralResult,
    PropertyMap,
    sort_configs_by_type,
)

BICR_COMPATIBLE = "nordic,nrf-bicr"


# Registers that must be set as unconstrained
BICR_UNCONSTRAINED = (
    "lfosc_lfxoconfig_0",
    "hfxo_startuptime_0",
    "hfxo_config_0",
)


class SvdEnum:
    """Helper class for constructing commonly used SVD enums."""

    UNCONFIGURED = "Unconfigured"

    @classmethod
    def int_or_unconfigured(cls, val: Optional[int]) -> Union[int, str]:
        return val if isinstance(val, int) else cls.UNCONFIGURED

    INTERNAL = "Internal"
    EXTERNAL = "External"
    SHORTED = "Shorted"

    @classmethod
    def rail_connection(cls, rail: str) -> str:
        for connection in [cls.INTERNAL, cls.EXTERNAL, cls.SHORTED]:
            if rail.lower() == connection.lower():
                return connection
        return cls.UNCONFIGURED

    ENABLED = "Enabled"
    DISABLED = "Disabled"

    @classmethod
    def enabled_if(cls, is_enabled: bool) -> str:
        return cls.ENABLED if is_enabled else cls.DISABLED

    NOT_PRESENT = "NotPresent"
    PRESENT = "Present"

    @classmethod
    def present_if(cls, is_present: bool) -> str:
        return cls.PRESENT if is_present else cls.NOT_PRESENT

    LFXO = "LFXO"
    LFRC = "LFRC"
    SYNTH = "Synth"

    @classmethod
    def lfosc_source(cls, source: str) -> str:
        for clksrc in [cls.LFXO, cls.LFRC, cls.SYNTH]:
            if source.lower() == clksrc.lower():
                return clksrc
        return cls.UNCONFIGURED

    CRYSTAL = "Crystal"
    EXTSQUARE = "ExtSquare"
    EXTSINE = "ExtSine"

    @classmethod
    def clk_mode(cls, mode: str) -> str:
        for binding, clkmode in [
            ("crystal", cls.CRYSTAL),
            ("external-square", cls.EXTSQUARE),
            ("external-sine", cls.EXTSINE),
            ("disabled", cls.DISABLED),
        ]:
            if mode.lower() == binding:
                return clkmode
        return cls.UNCONFIGURED

    @classmethod
    def lfosc_accuracy(cls, accuracy: Optional[int]) -> str:
        supported_ppms = [20, 30, 50, 75, 100, 150, 250, 500]

        if accuracy is None:
            return cls.UNCONFIGURED

        if accuracy not in supported_ppms:
            raise ValueError(
                f"LF oscillator accuracy must be one of {supported_ppms}; got {accuracy}"
            )

        return f"{accuracy}ppm"

    @classmethod
    def capacitance(cls, value: Optional[int]) -> Union[int, str]:
        if value is None:
            return cls.UNCONFIGURED

        return cls.EXTERNAL if value == 0 else value

    DISCONNECTED = "Disconnected"
    EXTERNAL_1V8 = "External1V8"
    EXTERNAL_3V = "External3V"
    EXTERNAL_FULL = "ExternalFull"

    @classmethod
    def port_rail(cls, value: int):
        connections = [
            cls.DISCONNECTED,
            cls.SHORTED,
            cls.EXTERNAL_1V8,
            cls.EXTERNAL_3V,
            cls.EXTERNAL_FULL,
        ]
        try:
            return connections[value]
        except IndexError:
            return cls.UNCONFIGURED

    OHMS = "Ohms"

    @classmethod
    def port_drive(cls, ohms: int):
        return f"{cls.OHMS}{ohms}"


@dc.dataclass(frozen=True, order=True)
class PowerConfig(ExtractedConfig):
    """Power rail configuration."""

    vddao0v8: str
    vdd1v0: str
    vddrf1v0: str
    vddao1v8: str
    vddao5v0: str
    vddvs0v8: str
    inductor: bool

    def to_record(self) -> Record:
        record = Record(
            {
                "power_config_0": {
                    "VDDAO0V8": SvdEnum.rail_connection(self.vddao0v8),
                    "VDD1V0": SvdEnum.rail_connection(self.vdd1v0),
                    "VDDRF1V0": SvdEnum.rail_connection(self.vddrf1v0),
                    "VDDAO1V8": SvdEnum.rail_connection(self.vddao1v8),
                    "VDDAO5V0": SvdEnum.rail_connection(self.vddao5v0),
                    "VDDVS0V8": SvdEnum.rail_connection(self.vddvs0v8),
                    "INDUCTOR": SvdEnum.present_if(self.inductor),
                }
            }
        )

        return record


class IOPortRail(NamedTuple):
    """Helper tuple for IOPORT rail configurations."""

    port: int
    rail: int


class IOPortDrive(NamedTuple):
    """Helper tuple for IOPORT drive configurations."""

    port: int
    resistance: int


@dc.dataclass(frozen=True, order=True)
class IOPortConfig(ExtractedConfig):
    """GPIO port configuration."""

    rails: List[IOPortRail] = dc.field(default_factory=list)
    drives: List[IOPortDrive] = dc.field(default_factory=list)

    def to_record(self) -> Record:
        records = defaultdict(dict)

        reg_idx = lambda c: "1" if c.port > 7 else "0"

        for cfg in self.rails:
            records[f"ioport_power{reg_idx(cfg)}_0"].update(
                {
                    f"P{cfg.port}": SvdEnum.port_rail(cfg.rail),
                }
            )

        for cfg in self.drives:
            records[f"ioport_drivectrl{reg_idx(cfg)}_0"].update(
                {
                    f"P{cfg.port}": SvdEnum.port_drive(cfg.resistance),
                }
            )

        return Record(records)


class LFAutoCal(NamedTuple):
    """Helper tuple for low-frequency oscillator autocalibration configurations."""

    temp_interval: int
    temp_delta: int
    max_num_intervals: int
    enabled: bool


@dc.dataclass(frozen=True, order=True)
class LFConfig(ExtractedConfig):
    """Low-frequency oscillator configuration."""

    src: str
    mode: str
    accuracy: Optional[int] = None
    loadcap: Optional[int] = None
    startup: Optional[int] = None
    autocalibration: Optional[LFAutoCal] = None

    def to_record(self) -> Record:
        record = Record()

        if self.src:
            record.update(
                {
                    "lfosc_config_0": {"SRC": SvdEnum.lfosc_source(self.src)},
                }
            )

        if any(
            x and x is not None
            for x in [self.accuracy, self.mode, self.loadcap, self.startup]
        ):
            record.update(
                {
                    "lfosc_lfxoconfig_0": {
                        "ACCURACY": SvdEnum.lfosc_accuracy(self.accuracy),
                        "MODE": SvdEnum.clk_mode(self.mode),
                        "LOADCAP": SvdEnum.capacitance(self.loadcap),
                        "TIME": SvdEnum.int_or_unconfigured(self.startup),
                    }
                }
            )

        if self.autocalibration:
            record.update(
                {
                    "lfosc_lfrcautocalconfig_0": {
                        "TEMPINTERVAL": self.autocalibration.temp_interval,
                        "TEMPDELTA": self.autocalibration.temp_delta,
                        "INTERVALMAXNO": self.autocalibration.max_num_intervals,
                        "ENABLE": SvdEnum.enabled_if(self.autocalibration.enabled),
                    }
                }
            )

        return record


@dc.dataclass(frozen=True, order=True)
class HFConfig(ExtractedConfig):
    """High-frequency oscillator configuration."""

    mode: str
    loadcap: Optional[int] = None
    startup: Optional[int] = None

    def to_record(self) -> Record:
        record = Record()

        if any(x and x is not None for x in [self.mode, self.loadcap]):
            record.update(
                {
                    "hfxo_config_0": {
                        "MODE": SvdEnum.clk_mode(self.mode),
                        "LOADCAP": SvdEnum.capacitance(self.loadcap),
                    }
                }
            )

        if self.startup:
            record.update(
                {
                    "hfxo_startuptime_0": {
                        "TIME": SvdEnum.int_or_unconfigured(self.startup),
                    }
                }
            )

        return record


@dc.dataclass(frozen=True, order=True)
class TamperConfig(ExtractedConfig):
    """Tamper controller configuration."""

    switch_enabled: bool
    shield_channels: List[int] = dc.field(default_factory=list)

    def to_record(self) -> Record:
        record = Record()

        if self.switch_enabled:
            record.update(
                {
                    "tampc_tamperswitch_0": {
                        "TAMPERSWITCH": SvdEnum.ENABLED,
                    }
                }
            )

        if self.shield_channels:
            record.update(
                {
                    "tampc_activeshield_0": {
                        f"CHEN_{ch}": SvdEnum.ENABLED for ch in self.shield_channels
                    }
                }
            )

        return record


class InterpretedBICR(InterpretedPeripheral):
    """
    BICR-related information interpreted from the parsed devicetree nodes that are relevant to the
    peripheral's configurations.
    """

    def __init__(self, devicetree: edtlib.EDT):
        """
        Initialize the class attribute(s).

        :param devicetree: edtlib Devicetree object
        """
        super().__init__(devicetree=devicetree, name="BICR")

        nodes = (ParsedDTNode(n) for n in devicetree.nodes)
        matching_nodes = [n for n in nodes if n.compatibles_contain(BICR_COMPATIBLE)]
        if not matching_nodes:
            raise RuntimeError(f"No BICR node with compatible {BICR_COMPATIBLE} found.")
        if len(matching_nodes) != 1:
            raise RuntimeError(
                f"Expected exactly one BICR node with compatible {BICR_COMPATIBLE}. "
                f"Found {len(matching_nodes)} nodes: {matching_nodes}."
            )

        self._node: ParsedDTNode = matching_nodes[0]
        self._properties.update(self._node.properties)

    def __str__(self) -> str:
        """String representation of the class."""

        props = {
            "Properties": [str(v) for v in self.properties.values()],
            "Chosen": self.chosen,
        }
        return f"Parsed BICR:\n{pformat(props)}\n"

    @property
    def address(self) -> int:
        """
        Base address of the BICR as defined by the devicetree.
        """
        address = self._node.address
        if address is None:
            raise RuntimeError(f"Missing 'reg' property in BICR node {self._node}")
        return address

    @cached_property
    def power_props(self) -> PropertyMap:
        """
        Mapping of properties related to BICR power configurations.
        """
        properties = PropertyMap(
            {
                k: v
                for k, v in self.properties.items()
                if any(k.startswith(x) for x in ["power", "inductor"])
            }
        )
        return properties

    @cached_property
    def lfosc_props(self) -> PropertyMap:
        """
        Mapping of properties related to low-frequency oscillator configurations.
        """
        properties = PropertyMap(
            {k: v for k, v in self.properties.items() if k.startswith("lf")}
        )
        return properties

    @cached_property
    def hfxo_props(self) -> PropertyMap:
        """
        Mapping of properties related to High-Frequency Clock configurations.
        """
        properties = PropertyMap(
            {k: v for k, v in self.properties.items() if k.startswith("hfxo")}
        )
        return properties

    @cached_property
    def tampc_props(self) -> PropertyMap:
        """
        Mapping of properties related to Tamper Controller configurations.
        """
        properties = PropertyMap(
            {k: v for k, v in self.properties.items() if k.startswith("tamper")}
        )
        return properties

    @cached_property
    def ioport_props(self) -> PropertyMap:
        """
        Mapping of properties related to GPIO port configurations.
        """
        properties = PropertyMap(
            {k: v for k, v in self.properties.items() if k.startswith("ioport")}
        )
        return properties


def from_dts(devicetree: edtlib.EDT) -> PeripheralResult:
    """
    Gather information from parsed devicetree nodes on configurations that must set in BICR
    in order to be configured on boot, then map those records into a memory map of BICR
    contents.

    :param devicetree: Devicetree structure.

    :return: Meta-record of extracted BICR records from devicetree nodes.
    """

    configs: List[ExtractedConfig] = []
    records: List[Record] = []

    bicr = InterpretedBICR(devicetree)

    log_dev(f"Parsed DTS nodes:\n{str(bicr)}")
    log_dev("")

    configs.append(extract_power_configs(bicr.power_props))
    configs.append(extract_ioport_configs(bicr.ioport_props))
    configs.append(extract_lfosc_configs(bicr.lfosc_props))
    configs.append(extract_hfxo_configs(bicr.hfxo_props))
    configs.append(extract_tamper_configs(bicr.tampc_props))

    for config_list in sort_configs_by_type(configs):
        for config in config_list:
            records.append(config.to_record())

    log_dev(f"All DTS records for BICR (pre-filter):")
    log_dev(pformat(records))
    log_dev("")

    bicr_record = Record()
    for record in [x for x in records if x]:
        bicr_record.update(record)

    log_dbg(f"Interpreted DTS records for BICR:")
    log_vrb(pformat(bicr_record))
    log_vrb("")

    return PeripheralResult(base_address=bicr.address, record=bicr_record)


def extract_power_configs(
    properties: PropertyMap,
) -> PowerConfig:
    """
    Extract power-related BICR configurations.

    :param properties: Power properties kept interpreted from devicetree's BICR.

    :return: List of all extracted configurations.
    """

    config = PowerConfig(
        vdd1v0=properties.get_val("power-vdd1v0", ""),
        vddao0v8=properties.get_val("power-vddao0v8", ""),
        vddao1v8=properties.get_val("power-vddao1v8", ""),
        vddao5v0=properties.get_val("power-vddao5v0", ""),
        vddrf1v0=properties.get_val("power-vddrf1v0", ""),
        vddvs0v8=properties.get_val("power-vddvs0v8", ""),
        inductor=properties.get_val("inductor-present", False),
    )

    return config


def extract_ioport_configs(
    properties: PropertyMap,
) -> PowerConfig:
    """
    Extract GPIO port-related BICR configurations.

    :param properties: GPIO port properties kept interpreted from devicetree's BICR.

    :return: List of all extracted configurations.
    """

    config = IOPortConfig(
        rails=[
            IOPortRail(port=entry.controller.props["port"].val, rail=entry.data["rail"])
            for entry in properties.get_val("ioport-power-rails", [])
        ],
        drives=[
            IOPortDrive(
                port=entry.controller.props["port"].val,
                resistance=entry.data["resistance"],
            )
            for entry in properties.get_val("ioport-drivectrls", [])
        ],
    )

    return config


def extract_lfosc_configs(
    properties: PropertyMap,
) -> PowerConfig:
    """
    Extract low-frequency oscillator-related BICR configurations.

    :param properties: LFOSC properties kept interpreted from devicetree's BICR.

    :return: List of all extracted configurations.
    """

    autocal_array = properties.get_val("lfrc-autocalibration")

    config = LFConfig(
        src=properties.get_val("lfosc-src", ""),
        mode=properties.get_val("lfosc-mode", ""),
        accuracy=properties.get_val("lfosc-accuracy"),
        loadcap=properties.get_val("lfosc-loadcap"),
        startup=properties.get_val("lfosc-startup"),
        autocalibration=(
            LFAutoCal(
                enabled=True,
                temp_interval=autocal_array[0],
                temp_delta=autocal_array[1],
                max_num_intervals=autocal_array[2],
            )
            if autocal_array
            else None
        ),
    )

    return config


def extract_hfxo_configs(
    properties: PropertyMap,
) -> PowerConfig:
    """
    Extract high-frequency oscillator-related BICR configurations.

    :param properties: HFXO properties kept interpreted from devicetree's BICR.

    :return: List of all extracted configurations.
    """

    config = HFConfig(
        mode=properties.get_val("hfxo-mode", ""),
        loadcap=properties.get_val("hfxo-loadcap"),
        startup=properties.get_val("hfxo-startup"),
    )

    return config


def extract_tamper_configs(
    properties: PropertyMap,
) -> PowerConfig:
    """
    Extract tamper controller (TAMPC)-related BICR configurations.

    :param properties: TAMPC properties kept interpreted from devicetree's BICR.

    :return: List of all extracted configurations.
    """

    config = TamperConfig(
        switch_enabled=properties.get_val("tamperswitch-enable", False),
        shield_channels=properties.get_val("tamper-activeshield-channels", []),
    )

    return config



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
This module provides helper functions and wrappers around edtlib devicetree objects that emphasize
aspects that are relevant in a product context, such as memory region and peripheral ownership.
"""

from __future__ import annotations

import collections.abc
import re
from abc import ABC, abstractmethod
from functools import cached_property, total_ordering
from itertools import groupby
from pprint import pformat
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Union

from devicetree import edtlib

from .common import AddressOffset, AddressRegion, DomainID, ProcessorID, Record, log_err


# Regex for known partition compatibles, which affect how the address of a
# parsed devicetree node should be interpreted.
PARTITION_COMPAT_REGEX = re.compile(r"^(fixed|nordic,owned)-partitions$")


class PropertyMap(collections.abc.Mapping):
    """Read-only map of devicetree properties"""

    def __init__(self, properties: Dict[str, edtlib.Property]):
        self._proxy = MappingProxyType(properties)

    def get_val(self, name: str, default: Optional[Any] = None) -> Optional[Any]:
        """
        Get the value of a devicetree property defined in the property map.
        Returns a default value if the property does not exist.

        :param name: Property name
        :param default: Default value to return if the property does not exist

        :return: Property value
        """
        if name not in self._proxy:
            return default
        return self._proxy[name].val

    def __getitem__(self, name: str) -> edtlib.Property:
        return self._proxy.__getitem__(name)

    def __iter__(self) -> Iterator[str]:
        return self._proxy.__iter__()

    def __len__(self) -> int:
        return self._proxy.__len__()

    def __str__(self) -> str:
        return f"<PropertyMap 0x{hex(id(self))}: {pformat(dict(self))}>"


@total_ordering
class ParsedDTNode:
    def __init__(self, node: edtlib.Node):
        """Initialize a ParsedDTNode from a devicetree node."""
        self._node: edtlib.Node = node

    @cached_property
    def name(self) -> str:
        """Name of the node"""
        return self.node.name.split("@")[0]

    @property
    def node(self) -> edtlib.Node:
        """The underlying edtlib.Node wrapped by this instance"""
        return self._node

    @property
    def status(self) -> str:
        """Status of the node."""
        return self.node.status

    @property
    def enabled(self) -> bool:
        """Node has status='okay'"""
        return self.status == "okay"

    @property
    def reserved(self) -> bool:
        """Node has status='reserved'"""
        return self.status == "reserved"

    @cached_property
    def parent(self) -> Optional[ParsedDTNode]:
        """Parent node"""
        if not self.node.parent:
            return None
        return ParsedDTNode(self.node.parent)

    @cached_property
    def address(self) -> Optional[int]:
        """Absolute address of the node"""
        if not self.node.regs:
            return None

        addr = self.node.regs[0].addr
        parent = self.parent

        if parent and parent.compatibles_match(PARTITION_COMPAT_REGEX):
            gparent = parent.node.parent
            if gparent:
                if not gparent.regs:
                    raise RuntimeError(f"Missing 'reg' property in node {gparent.path}")
                addr += gparent.regs[0].addr

        return addr

    @property
    def length(self) -> Optional[int]:
        """Address space size of the node"""
        if not self.node.regs:
            return None
        return self.node.regs[0].size

    @property
    def path(self) -> str:
        """Devicetree node path of the node"""
        return self.node.path

    @cached_property
    def secure(self) -> Optional[bool]:
        """
        Security attribute of the node, determined solely by its bus. In Trustzone-enabled systems,
        this is determined by the security bit in the node's absolute address.

        This attribute does not represent any other context of security that could relate to the
        node, such as properties denoting secure permissions, memory, etc.
        """
        if self.address is None:
            return None

        address = self.node.bus_node.unit_addr if self.node.bus_node else self.address

        return bool(address & (1 << AddressOffset.SECURITY))

    @cached_property
    def properties(self) -> PropertyMap:
        """Devicetree properties defined on the node"""
        return PropertyMap(self.node.props)

    @property
    def labels(self) -> List[str]:
        """Labels"""
        return self.node.labels

    @property
    def compatibles(self) -> List[str]:
        """Compatible property strings if one or more exist, else empty."""
        return self.node.compats

    @property
    def domain(self) -> Optional[DomainID]:
        """Domain ID associated to the node's bus address."""
        if self.address is None:
            return None
        return DomainID.from_address(self.address)

    @property
    def region(self) -> Optional[AddressRegion]:
        """Address region associated to the node's bus address."""
        if self.address is None:
            return None
        return AddressRegion.from_address(self.address)

    @property
    def pinctrls(self) -> List[edtlib.PinCtrl]:
        """Pinctrl configurations referenced by the node"""
        return self.node.pinctrls

    def __str__(self) -> str:
        """String representation of the class."""

        attrs = {
            "Properties": [str(v) for v in self.properties.values()],
            "Nodelabels": self.labels,
            "Path": self.path,
        }
        address_str = "" if self.address is None else f"@{hex(self.address)}"
        domain_str = (
            ""
            if self.domain is None
            else f" - {self.domain.name} Domain ({self.domain})"
        )
        return (
            f"ParsedDTNode {self.name}{address_str}{domain_str}: " f" {pformat(attrs)}"
        )

    def label_string(self, upper: bool = True) -> str:
        """
        Helper for returning an instance string of the node.

        :return: Formatted instance string.
        """
        label = (
            f" ({self.labels[0].upper() if upper else self.labels[0]})"
            if self.labels
            else ""
        )
        return f"0x{self.address:08x}{label}"

    def compatibles_contain(self, substring: str) -> bool:
        """
        Helper method for checking the existence of a substring in all of the node's compatible
        strings.

        :param substring: Compatible substring to check against

        :return: True if any compatible contains the substring, False otherwise.
        """

        return any(substring in compatible for compatible in self.compatibles)

    def compatibles_match(self, pattern: Union[str, re.Pattern]) -> bool:
        """
        Helper method for checking if a regular expression matches any of the node's compatible
        strings.

        :param pattern: Pattern to check against

        :return: True if any compatible matches the pattern, False otherwise.
        """

        return any(
            re.match(pattern, compatible) is not None for compatible in self.compatibles
        )

    def __eq__(self, other: object) -> bool:
        """Equality operator based on the node's dep_ordinal value"""

        if not isinstance(other, ParsedDTNode):
            return False
        # edtlib.Node.dep_ordinal uniquely identifies a node in a given EDT
        return (
            self._node.edt is other._node.edt
            and self._node.dep_ordinal == other._node.dep_ordinal
        )

    def __lt__(self, other: object) -> bool:
        """Less than operator based on the node's address, size and name (in that order)"""

        if not isinstance(other, ParsedDTNode):
            return NotImplemented

        return (self.address, self.length, self.name) < (
            other.address,
            other.length,
            other.name,
        )

    def __hash__(self) -> int:
        """Hash operator based on the node's dep_ordinal value"""

        return hash((id(self._node.edt), self._node.dep_ordinal))


class InterpretedPeripheral(ABC):
    """
    Abstract base class of a peripheral that is interpreted from devicetree configurations.
    """

    def __init__(self, devicetree: edtlib.EDT, name: str):
        """
        Initialize the class attribute(s).

        :param devicetree: edtlib Devicetree object
        :param name: Peripheral name
        """
        self._name = name.upper()
        self._properties: Dict[str, edtlib.Property] = {}
        self._chosen_nodes: Dict[str, ParsedDTNode] = {
            name: ParsedDTNode(node) for name, node in devicetree.chosen_nodes.items()
        }

    @property
    def name(self) -> str:
        """
        Name of the interpreted peripheral.
        """
        return self._name

    @property
    @abstractmethod
    def address(self) -> int:
        """
        Base address of the peripheral.
        """
        ...

    @cached_property
    def properties(self) -> PropertyMap:
        """
        Peripheral properties defined for the peripheral in devicetree.
        """
        return PropertyMap(self._properties)

    @property
    def chosen(self) -> Dict[str, ParsedDTNode]:
        """
        Chosen nodes in the devicetree.
        """
        return self._chosen_nodes

    def __str__(self) -> str:
        """String representation of the class."""

        props = {
            "Properties": [str(v) for v in self.properties.values()],
            "Chosen": self.chosen,
        }
        nodes = {k.name: [str(n) for n in v] for k, v in self._resource_nodes.items()}
        return f"Parsed {self.name}:\n{pformat(props)}\n{pformat(nodes)}"


class ProcessorInfo(NamedTuple):
    """
    Helper tuple for containing relevant information about the parsed CPU.
    """

    cpu_type: str
    cpu_id: ProcessorID

    def __repr__(self) -> str:
        """
        Basic class representation
        """
        return f"<ProcessorInfo at {hex(id(self))} cpu_type={self.cpu_type}, cpu_id={self.cpu_id}>"

    def __str__(self) -> str:
        """
        String representation of the processor info.
        """
        make, model = (x.capitalize() for x in self.cpu_type.split(",", maxsplit=1))
        return f"Processor {self.cpu_id} - {make} {model}"


class PeripheralResult(NamedTuple):
    """
    Information about a peripheral that is extracted from a devicetree.
    """

    base_address: int
    record: Record


def dt_processor_info(devicetree: edtlib.EDT) -> ProcessorInfo:
    """
    Get processor information from a domain's devicetree.

    :param devicetree: Devicetree

    :return: Tuple of processor information of the devicetree's domain
    """
    cpus = [
        node
        for node in devicetree.get_node("/cpus").children.values()
        if node.name.startswith("cpu@")
    ]
    if len(cpus) != 1:
        raise RuntimeError(
            f"Expected exactly 1 'cpu' node, but devicetree contained {len(cpus)} nodes"
        )

    cpu = cpus[0]
    try:
        compatible = cpu.compats[0]
    except IndexError:
        raise RuntimeError("Devicetree 'cpu' node has no compatible")

    processor_id = ProcessorID.from_value(cpu.regs[0].addr)
    if processor_id is None:
        raise RuntimeError(
            f"Devicetree 'cpu' node has invalid Processor ID {cpu.regs[0].addr}"
        )

    return ProcessorInfo(compatible, processor_id)


class NodeChannels:
    """Helper class for encapsulating channel resources for a node under a common object."""

    __slots__ = ["_owned", "_child", "_nonsecure", "_all", "_source", "_sink"]

    def __init__(
        self,
        owned: List[int] = [],
        child: List[int] = [],
        nonsecure: List[int] = [],
        source: List[int] = [],
        sink: List[int] = [],
    ):
        self._owned = owned
        self._child = child
        self._nonsecure = nonsecure
        self._all = list(set(self._owned) | set(self._child))
        self._source = source
        self._sink = sink

    @property
    def owned(self) -> List[int]:
        """Privately owned channels by the node."""
        return self._owned

    @property
    def child(self) -> List[int]:
        """Channels owned by the node on behalf of a child subprocessor."""
        return self._child

    @property
    def nonsecure(self) -> List[int]:
        """Channels owned by the node that are nonsecure."""
        return self._nonsecure

    @property
    def all(self) -> List[int]:
        """All channels owned by the node."""
        return self._all

    @property
    def source(self) -> List[int]:
        """Channels owned by the node that are configured as sources."""
        return self._source

    @property
    def sink(self) -> List[int]:
        """Channels owned by the node that are configured as sinks."""
        return self._sink

    def __str__(self) -> str:
        """String representation of the class."""
        if self.child or self.nonsecure or self.source or self.sink:
            channels = {
                f"{'Self-o' if self.child else 'O'}wned": self.owned,
                "Child-owned": self.child,
                "Nonsecure": self.nonsecure,
                "Sources": self.source,
                "Sinks": self.sink,
            }
            info_str = "\n" + "\n".join(
                [f"\t\t{k}: {v}" for k, v in channels.items() if v]
            )
        else:
            info_str = f" {self.owned}"

        return f"Channels:{info_str}"


class IndexedByInstance(NamedTuple):
    """Index by instance address"""

    # Instance node to index the configuration by
    instance: ParsedDTNode


class IndexedByPosition(NamedTuple):
    """Index by array position"""

    # Number of array slots taken up by one configuration object
    num_slots: int


# Methods for indexing configuration objects
IndexedBy = Union[IndexedByInstance, IndexedByPosition]


class ExtractedConfig(ABC):
    """Base class for extracted configuration objects."""

    @property
    def indexed_by(self) -> Optional[IndexedBy]:
        """
        If the configuration is meant for an array register, return metadata describing how to
        assign an array index to the  configuration.
        If None, the configuration is not meant for an array register and does not require an index.
        """
        return None

    @abstractmethod
    def to_record(self, *, index: int, **kwargs) -> Record:
        """
        Convert object to a register Record representation.
        The return value contains both the Record and the number of register slots taken up by the
        Record. The number of slots can be used to assign the Record to the correct register index.

        :param index: Index of the register to assign the Record to.
        :param kwargs: Unspecified keyword arguments to use in constructing the Record.
        :return: A tuple of the Record and the number of registers represented in the Record.
        """
        raise NotImplementedError


def sort_configs_by_type(configs: List[ExtractedConfig]) -> List[List[ExtractedConfig]]:
    """
    Sort configs by type, then group by type and sort each group internally.
    This ensures that the order of the configs in the output record is consistent.
    """
    type_sorted_configs = sorted(configs, key=lambda c: type(c).__name__)

    return [sorted(c) for _, c in groupby(type_sorted_configs, key=type) if c]
